/*
  iowlcd.c
 
  a library to control a LCD 
  connected to the IO-Warrior from www.codemercs.com
  
  written by Christian Vogelgsang <chris@lallafa.de>
  under the GNU Public License Version 2
 
  $Id: iowlcd.c,v 1.2 2004/05/01 19:49:48 cnvogelg Exp $
*/

#include <string.h>
#include "iowlcd.h"

/* define some IOW constants */
#define iowVendor  0x07c0   /* CodeMercs */
#define iowProd40  0x1500   /* IOW40 */
#define iowProd24  0x1501   /* IOW24 */
#define iowTimeout   1000

#define USB_REQ_SET_REPORT 0x09
#define USB_REQ_GET_REPORT 0x01

/* ----- iow_init ---------------------------------------------------------- */
/* returns the number of IOW found */

static int iow_num_devs;
static struct usb_device **iow_devs;

int iow_init()
{
  struct usb_bus *busses;
  struct usb_bus *bus;
  struct usb_device *device;

  usb_init();
  usb_find_busses();
  usb_find_devices();

  /* now scan through busses to find matching adapters */
  int num = 0;
  busses = usb_get_busses();
  for(bus=busses;bus!=NULL;bus=bus->next) {
    for(device=bus->devices;device!=NULL;device=device->next) {
      /* codemercs ? */
      if(device->descriptor.idVendor==iowVendor) {
        if((device->descriptor.idProduct==iowProd24)||
           (device->descriptor.idProduct==iowProd40)) {
          num++;
        }
      }
    }
  }
  
  /* found some warriors */
  if(num>0) {
    struct usb_device **devs;
    
    iow_devs = (struct usb_device **)malloc(sizeof(struct usb_device *)*num);
    devs = iow_devs;
    
    /* scan again and store devices */
    for(bus=busses;bus!=NULL;bus=bus->next) {
      for(device=bus->devices;device!=NULL;device=device->next) {
        /* codemercs ? */
        if(device->descriptor.idVendor==iowVendor) {
          if((device->descriptor.idProduct==iowProd24)||
             (device->descriptor.idProduct==iowProd40)) {
            *devs = device;
            devs++;
          }
        }
      }
    }
  }
  iow_num_devs = num;
  return num;
}

/* ----- iow_free ---------------------------------------------------------- */

void iow_free()
{
  free(iow_devs);
  iow_devs = 0;
  iow_num_devs = 0;
}

/* ----- iow_open ---------------------------------------------------------- */

usb_dev_handle *iow_open(int id)
{
  usb_dev_handle *udh;
  
  /* valid id? */
  if((id>=iow_num_devs)||(id<0))
    return NULL;
  
  /* open USB device */
  udh = usb_open(iow_devs[id]);
  if(udh==0)
    return NULL;
  
  /* set configuration */
  if(usb_set_configuration(udh,1)<0) {
    usb_close(udh);
    return NULL;
  }
    
  /* claim interface 1 */
  if(usb_claim_interface(udh,1)<0) {
    usb_close(udh);
    return NULL;
  }
  
  return udh;
}

/* ----- iow_close --------------------------------------------------------- */

void iow_close(usb_dev_handle *udh)
{
  /* release interface 1 */
  usb_release_interface(udh,1);
  
  /* close USB */
  usb_close(udh);
}

/* ----- iow_wcmd ---------------------------------------------------------- */
/* write a set report to interface 1 of warrior */
static int iow_wcmd(usb_dev_handle *udh,unsigned char report[8])
{
  return (usb_control_msg(udh,USB_DT_HID,USB_REQ_SET_REPORT,0,1,
                          report,8,iowTimeout)==8) ? IOW_OK : IOW_ERROR;
}

/* ----- iow_rcmd ---------------------------------------------------------- */
static int iow_rcmd(usb_dev_handle *udh,unsigned char report[8])
{
  /* 2 = interrupt endpoint for command read back on IOW */
  return (usb_interrupt_read(udh,2,report,8,iowTimeout)==8) ? IOW_OK : IOW_ERROR;
}

/* ========== IOW LCD Tools ================================================ */

int iowlcd_enable(usb_dev_handle *udh)
{
  unsigned char lcd_cmd[8] = { 0x04, 0x01, 0, 0, 0, 0, 0, 0 };
  int res = iow_wcmd(udh,lcd_cmd);
  usleep(30000); /* wait for 30ms */
  return res;
}

int iowlcd_disable(usb_dev_handle *udh)
{
  unsigned char lcd_cmd[8] = { 0x04, 0x00, 0, 0, 0, 0, 0, 0 };
  int res = iow_wcmd(udh,lcd_cmd);
  usleep(30000);
  return res;
}

/* status: 0x80=busy bit, 0x7f-> addr */
int iowlcd_read_status(usb_dev_handle *udh)
{
  unsigned char lcd_cmd[8] = { 0x06, 1, 0, 0, 0, 0, 0, 0 };
  
  if(iow_wcmd(udh,lcd_cmd)==IOW_ERROR)
    return IOW_ERROR;
  if(iow_rcmd(udh,lcd_cmd)==IOW_ERROR)
    return IOW_ERROR;
  
  /* cmd id = 0x06 */
  if(lcd_cmd[0]!=6)
    return IOW_ERROR;
  /* len must be 1 */
  if(lcd_cmd[1]!=1)
    return IOW_ERROR;

  return lcd_cmd[2];
}

/* LCD commands */

int iowlcd_display_clear(usb_dev_handle *udh)
{
  unsigned char lcd_cmd[8] = { 0x05, 1, 0x01, 0, 0, 0, 0, 0 };
  int res = iow_wcmd(udh,lcd_cmd);
  usleep(3000); /* 3ms */
  return res;
}

int iowlcd_return_home(usb_dev_handle *udh)
{
  unsigned char lcd_cmd[8] = { 0x05, 1, 0x03, 0, 0, 0, 0, 0 };
  int res = iow_wcmd(udh,lcd_cmd);
  usleep(3000); /* 3ms */
  return res;
}

int iowlcd_entry_mode_set(usb_dev_handle *udh,int inc,int shift)
{
  unsigned char lcd_cmd[8] = { 0x05, 1, 0x04, 0, 0, 0, 0, 0 };
  if(inc)   lcd_cmd[2] |= 0x02;
  if(shift) lcd_cmd[2] |= 0x01;
  return iow_wcmd(udh,lcd_cmd);
}

int iowlcd_display_on_off(usb_dev_handle *udh,int display,int cursor,int blink)
{
  unsigned char lcd_cmd[8] = { 0x05, 1, 0x08, 0, 0, 0, 0, 0 };
  if(display) lcd_cmd[2] |= 0x04;
  if(cursor)  lcd_cmd[2] |= 0x02;
  if(blink)   lcd_cmd[2] |= 0x01;
  return iow_wcmd(udh,lcd_cmd);  
}

int iowlcd_shift(usb_dev_handle *udh,int display,int right)
{
  unsigned char lcd_cmd[8] = { 0x05, 1, 0x10, 0, 0, 0, 0, 0 };
  if(display) lcd_cmd[2] |= 0x08;
  if(right)   lcd_cmd[2] |= 0x04;
  return iow_wcmd(udh,lcd_cmd);  
}

int iowlcd_set_function(usb_dev_handle *udh,int eight_bit,int two_line,int ten_dots)
{
  unsigned char lcd_cmd[8] = { 0x05, 1, 0x20, 0, 0, 0, 0, 0 };
  if(eight_bit) lcd_cmd[2] |= 0x10;
  if(two_line)  lcd_cmd[2] |= 0x08;
  if(ten_dots)  lcd_cmd[2] |= 0x04;
  return iow_wcmd(udh,lcd_cmd);  
}

int iowlcd_set_cgram_addr(usb_dev_handle *udh,int addr)
{
  unsigned char lcd_cmd[8] = { 0x05, 1, 0x40, 0, 0, 0, 0, 0 };
  lcd_cmd[2] |= (addr & 0x3f);
  return iow_wcmd(udh,lcd_cmd);  
}

int iowlcd_set_ddram_addr(usb_dev_handle *udh,int addr)
{
  unsigned char lcd_cmd[8] = { 0x05, 1, 0x80, 0, 0, 0, 0, 0 };
  lcd_cmd[2] |= (addr & 0x7f);
  return iow_wcmd(udh,lcd_cmd);  
}

int iowlcd_write_data(usb_dev_handle *udh,int len,unsigned char *data)
{
  unsigned char lcd_cmd[8] = { 0x05, 0x80, 0, 0, 0, 0, 0, 0 };
  unsigned char *ptr = data;
  int num_blk, last_blk, i;
  num_blk  = len / 6;
  last_blk = len % 6;
  
  /* write 6 data byte reports */
  for(i=0;i<num_blk;i++) {
    lcd_cmd[1] = 0x86;
    memcpy(&lcd_cmd[2],ptr,6);
    if(iow_wcmd(udh,lcd_cmd)==IOW_ERROR)
      return ptr - data;
    ptr += 6;
  }
  
  /* last block */
  if(last_blk>0) {
    lcd_cmd[1] = 0x80 | last_blk;
    memcpy(&lcd_cmd[2],ptr,last_blk);
    if(iow_wcmd(udh,lcd_cmd)==IOW_ERROR)
      return ptr - data;
  }
  
  return len;
}

int iowlcd_read_data(usb_dev_handle *udh,int len,unsigned char *data)
{
  unsigned char lcd_cmd[8] = { 0x06, 0, 0, 0, 0, 0, 0, 0 };
  unsigned char *ptr = data;
  int num_blk, last_blk, i;
  
  len = len & 0x3f;
  num_blk  = len / 6;
  last_blk = len % 6;
  
  /* full blocks */
  for(i=0;i<num_blk;i++) {
    /* write command */
    lcd_cmd[1] = 0x86;
    if(iow_wcmd(udh,lcd_cmd)==IOW_ERROR)
      return ptr - data;
    /* read command */
    if(iow_rcmd(udh,lcd_cmd)==IOW_ERROR)
      return ptr - data;
    memcpy(ptr,&lcd_cmd[2],6);
    ptr += 6;
  }

  /* last block */
  if(last_blk>0) {
    /* write command */
    lcd_cmd[1] = 0x80 | last_blk;
    if(iow_wcmd(udh,lcd_cmd)==IOW_ERROR)
      return ptr-data;
    if(iow_rcmd(udh,lcd_cmd)==IOW_ERROR)
      return ptr-data;
    memcpy(ptr,&lcd_cmd[2],last_blk);
  }
  
  return len;
}

/* ========== IOW LCD Tools ================================================ */

int iowlcd_setup(usb_dev_handle *udh)
{
  /* enable LCD in IOW */
  if(iowlcd_enable(udh)==IOW_ERROR)
    return IOW_ERROR;
  /* enable 8bit transfer mode */
  if(iowlcd_set_function(udh,1,1,0)==IOW_ERROR)
    return IOW_ERROR;
  /* clear screen */
  if(iowlcd_display_clear(udh)==IOW_ERROR)
    return IOW_ERROR;
  /* enable display, disable cursor+blinking */
  if(iowlcd_display_on_off(udh,1,0,0)==IOW_ERROR)
    return IOW_ERROR;
  
  return IOW_OK;
}

int iowlcd_shutdown(usb_dev_handle *udh)
{
  /* display off */
  if(iowlcd_display_on_off(udh,0,0,0)==IOW_ERROR)
    return IOW_ERROR;
  /* IOW leave LCD mode */
  if(iowlcd_disable(udh)==IOW_ERROR)
    return IOW_ERROR;
  
  return IOW_OK;
}

int iowlcd_set_pos(usb_dev_handle *udh,int x,int y)
{
  const unsigned char lineOff[4] = { 0x00,0x40,0x14,0x54 };
  unsigned char addr = lineOff[y] + x;
  return iowlcd_set_ddram_addr(udh,addr);
}

int iowlcd_set_text(usb_dev_handle *udh,int x,int y,int len,unsigned char *data)
{
  if(iowlcd_set_pos(udh,x,y)==IOW_ERROR)
    return IOW_ERROR;
  return iowlcd_write_data(udh,len,data);
}

int iowlcd_get_text(usb_dev_handle *udh,int x,int y,int len,unsigned char *data)
{
  if(iowlcd_set_pos(udh,x,y)==IOW_ERROR)
    return IOW_ERROR;
  return iowlcd_read_data(udh,len,data);
}

int iowlcd_load_chars(usb_dev_handle *udh,int offset,int num,unsigned char *bits)
{
  if(iowlcd_set_cgram_addr(udh,offset*8)==IOW_ERROR)
    return IOW_ERROR;
  return iowlcd_write_data(udh,num*8,bits);
}

/* bar drawing */

int iowlcd_load_vbar_chars(usb_dev_handle *udh)
{
  unsigned char bits[8*7];
  int i,j;
  unsigned char *ptr = bits;
  
  /* create vbar chars */
  for(i=0;i<7;i++) {
    for(j=7;j>=0;j--) {
      if(j<=i)
        *ptr = 0xff;
      else
        *ptr = 0x00;
      ptr++;
    }
  }
  
  return iowlcd_load_chars(udh,0,7,bits);
}

int iowlcd_draw_vbar(usb_dev_handle *udh,int x,int y,int h,int val)
{
  unsigned char bar;
  int i,char_min,char_max;
  
  /* build up bars row by row to use fastest access */
  char_min = 1;
  char_max = 7;
  for(i=0;i<h;i++) {
    int line = y + h - 1 - i;
    /* check all values in this line */
    if(val<char_min)
      bar = 32;
    else if(val>char_max)
      bar = 255;
    else
      bar = (unsigned char)(val - char_min);
    if(iowlcd_set_text(udh,x,line,1,&bar)==IOW_ERROR)
      return IOW_ERROR;
    char_min += 8;
    char_max += 8;
  }
  return IOW_OK;
}

int iowlcd_draw_vbars(usb_dev_handle *udh,int x,int y,int h,int num,int *vals)
{
  unsigned char bar[20];
  int i,j,char_min,char_max;
  
  /* build up bars row by row to use fastest access */
  char_min = 1;
  char_max = 7;
  for(i=0;i<h;i++) {
    int line = y + h - 1 - i;
    /* check all values in this line */
    for(j=0;j<num;j++) {
      int val = vals[j];
      if(val<char_min)
        bar[j] = 32;
      else if(val>char_max)
        bar[j] = 255;
      else
        bar[j] = (unsigned char)(val - char_min);
    }
    if(iowlcd_set_text(udh,x,line,num,bar)==IOW_ERROR)
      return IOW_ERROR;
    char_min += 8;
    char_max += 8;
  }
  return IOW_OK;
}

int iowlcd_load_hbar_chars(usb_dev_handle *udh)
{
  unsigned char bits[8*4];
  int i,j;
  unsigned char *ptr = bits;
  unsigned char val[4] = { 0x10,0x18,0x1c,0x1e };
  
  /* create vbar chars */
  for(i=0;i<4;i++) {
    for(j=0;j<8;j++) {
      *ptr = val[i];
      ptr++;
    }
  }
  
  return iowlcd_load_chars(udh,0,4,bits);
}

int iowlcd_draw_hbar(usb_dev_handle *udh,int x,int y,int w,int val)
{
  unsigned char bar[20];
  int full = val / 5;
  int rest = val % 5;
  int i;
  
  /* full bar */
  if(full>=w) {
    memset(bar,255,w);
    return iowlcd_set_text(udh,x,y,w,bar);
  }
  /* empty bar */
  else if(val==0) {
    memset(bar,32,w);
    return iowlcd_set_text(udh,x,y,w,bar);
  }

  /* partial bar */
  memset(bar,32,w);
  for(i=0;i<full;i++)
    bar[i] = 255;
  if(rest>0) {
    bar[full] = rest-1;
  }
  return iowlcd_set_text(udh,x,y,w,bar);
}

int iowlcd_draw_hbars(usb_dev_handle *udh,int x,int y,int w,int num,int *vals)
{
  int i;
  
  for(i=0;i<num;i++) {
    if(iowlcd_draw_hbar(udh,x,y+i,w,vals[i])==IOW_ERROR)
      return IOW_ERROR;
  }
  return IOW_OK;
}
