#!/usr/bin/perl -w
#
# iowlcdml.pl
#
# display pages written in LCDML
# a very simple markup language for LCD status display
#
# written by Christian Vogelgsang <chris@vogelgsang.org>
# under the GNU Public License V2

use iowlcd;
use XML::Simple;
use Time::HiRes qw( sleep );
use Data::Dumper;
use Getopt::Std;

# parse parameter
our($opt_d,$opt_f);
getopts('d:f:');
$opt_d = 0 unless defined($opt_d);
$opt_f = "lcdml.xml" unless defined($opt_f);

# read in the XML file
my $xml = XMLin($opt_f,ForceContent => 1, KeyAttr => {});
die "can't read XML file" if(!defined($xml));
printDebug(7,Dumper($xml));

# check display spec
$loop   = $xml->{'loop'};
$device = $xml->{'device'};
$size   = $xml->{'size'};
$loop   = 0 unless(defined($loop));
$device = 0 unless(defined($device));
$size   = "20,4" unless(defined($size));
($width,$height) = split(/,/,$size);
printDebug(1,"display device=$device size=$width,$height loop=$loop\n");

# find pages
$pages = $xml->{'page'};
die "No Pages defined!!" if(!defined($pages));
# more pages?
push_one_or_more(\@my_pages,$pages);
printDebug(1,"pages #" . scalar(@my_pages) . "\n");

# --- init iowlcd ---
$num_warrior = iowlcd::iow_init();
if($device >= $num_warrior) {
  print STDERR "device $device not available (only $num_warrior found)!\n";
  exit 1;
}
$udh = iowlcd::iow_open($device);
if(iowlcd::iowlcd_setup($udh)==$iowlcd::IOW_ERROR) {
  print STDERR "LCD setup failed!\n";
  exit 1;
}

# empty my string table
%lcdml_strings = ();

# main loop
if($loop == -1) {
  while(1) {
    foreach(@my_pages) { display_page($_); }
  }
} else {
  for(0..$loop) {
    foreach(@my_pages) { display_page($_); }
  }
}

# ----- shut down LCD -----
iowlcd::iowlcd_shutdown($udh);
iowlcd::iow_close($udh);
iowlcd::iow_free();
exit 0;

# ---------- display a page on the LCD ----------------------------------------
sub display_page {
  my($page) = shift;
  
  # extract options: duration and clear
  my($duration) = $page->{'duration'};
  my($timestep) = $page->{'timestep'};
  my($clear)    = $page->{'clear'};
  $duration = 5 unless(defined($duration));
  $timestep = $duration unless(defined($timestep));
  $clear    = 1 unless(defined($clear));
  
  # find regions
  my($regions)  = $page->{'region'};
  my(@my_regions);
  push_one_or_more(\@my_regions,$regions);
  printDebug(1,"page: duration=$duration timestep=$timestep clear=$clear regions=#" 
    . scalar(@my_regions) . "\n");
  
  # clear display
  iowlcd::iowlcd_display_clear($udh) if($clear);
  
  # the display loop of the current page
  my($max_steps) = int($duration / $timestep + 0.5);
  for($t=0;$t<$max_steps;$t++) {
    printDebug(2,"  timestep #" . $t . "\n");
    # draw region
    foreach(@my_regions) {
      draw_region($t,$_);
    }
    sleep($timestep);
  }
}

# ---------- draw a region ----------------------------------------------------
sub draw_region {
  my($time)   = shift;
  my($region) = shift;
  
  # fetch parameters
  my($update) = $region->{'update'};
  my($size)   = $region->{'size'};
  my($pos)    = $region->{'pos'};
  $update = 0 unless(defined($update));
  $size = "$width,$height" unless(defined($size));
  $pos  = "0,0" unless(defined($pos));
  my($x,$y,$w,$h);
  ($w,$h) = split(/,/,$size);
  ($x,$y) = split(/,/,$pos);
  my(@geo);
  @geo = ($x,$y,$w,$h);

  # is the right time?
  if($update==0) {
    return if($time!=0);
  } else {
    return if(($time % $update)!=0); 
  }
  printDebug(3,"    region update=$update pos=$x,$y size=$w,$h\n");
  
  # ----- handle elements of the region -----
  # 'string'
  my($st) = $region->{'string'};
  if(defined($st)) {
    my(@str);
    push_one_or_more(\@str,$st);
    foreach(@str) {
      set_string($_);
    }
  }
  # 'text'
  my($tx) = $region->{'text'};
  if(defined($tx)) {
    my(@text);
    push_one_or_more(\@text,$tx);
    foreach(@text) {
      draw_text($_,\@geo);
    }
  }
  # 'bars'  
  my($ba) = $region->{'bars'};
  if(defined($ba)) {
    my(@bars);
    push_one_or_more(\@bars,$ba);
    my($do_load) = ($time == 0);
    foreach(@bars) {
      draw_bars($_,\@geo,$do_load);
      $do_load = 0;
    }
  }
}

# ---------- draw text --------------------------------------------------------
sub draw_text {
  my($text,$geo) = @_;
  my($string) = parse_content($text);
  my($x,$y) = parse_pos($text);
  
  # maximum width
  my($allow_w) = $$geo[2] - $x;
  
  # multiple rows?
  my(@rows) = split(/\n/,$string);
  my($xp) = $x + $$geo[0];
  my($yp) = $y + $$geo[1];
  foreach(@rows) {
    # need to clamp row?
    if(length($_)>$allow_w) {
      $_ = substr($_,0,$allow_w);
    }  
    printDebug(4,"      text pos=$xp,$yp -> '$_'\n");
    iowlcd::iowlcd_set_text($udh,$xp,$yp,$_);
    $yp++;
    # too many rows?
    last if($yp>=($$geo[1]+$$geo[3]));
  }
}

# ---------- draw bars --------------------------------------------------------
sub draw_bars {
  my($bars,$geo,$do_load) = @_;
  
  # fetch pos
  my($x,$y) = parse_pos($bars);
  
  # fetch direction
  my($dir) = $bars->{'dir'};
  $dir = 'h' unless defined($dir);

  # load chars
  if($do_load) {
    if($dir eq 'h') {
      iowlcd::iowlcd_load_hbar_chars($udh);
    } else {
      iowlcd::iowlcd_load_vbar_chars($udh);
    }
  }
    
  # fetch size
  my($size) = $bars->{'size'};
  if(!defined($size)) {
    if($dir eq 'h') {
      $size = $$geo[2] - $x;
    } else {
      $size = $$geo[3] - $y;
    }
  }

  # calc maximum val
  my($maxval);
  if($dir eq 'h') {
    $maxval = $size * $iowlcd::IOWLCD_HBAR_UNITS;
  } else {
    $maxval = $size * $iowlcd::IOWLCD_VBAR_UNITS;
  }
  
  my($xp) = $x + $$geo[0];
  my($yp) = $y + $$geo[1];
  printDebug(4,"      bar pos=$xp,$yp size=$size dir=$dir maxval=$maxval\n");
  
  # retrieve all embedded bars
  my($bar) = $bars->{'bar'};
  my(@bars);
  push_one_or_more(\@bars,$bar);
  
  # fetch values from bars
  my(@vals);
  foreach(@bars) {
    my($value) = parse_content($_);
    
    # specify either offset,scale or min,max
    my($offset) = $_->{'offset'};
    my($scale) = $_->{'scale'};
    my($min) = $_->{'min'};
    my($max) = $_->{'max'};
    my($result);
    if(defined($offset)||defined($scale)) {
      $offset = 0 unless defined($offset);
      $scale = 1 unless defined($scale);
      $result = $offset + $value * $scale;
    } else {
      $min = 0 unless defined($min);
      $max = $maxval unless defined($max);
      $result = ($value - $min) * $maxval / ($max - $min);
    }

    # clamp
    $result = int($result + 0.5);
    $result = 0 if($result < 0);
    $result = $maxval if($result > $maxval);
    
    printDebug(4,"\t        bar value=$value -> result=$result\n");
    push @vals,$result;
  }

  # draw bars
  if($dir eq 'h') {
    iowlcd::iowlcd_draw_hbars($udh,$xp,$yp,$size,\@vals);
  } else {
    iowlcd::iowlcd_draw_vbars($udh,$xp,$yp,$size,\@vals);
  }
}

# ---------- string handling --------------------------------------------------
sub set_string {
  my($obj) = shift;
  my($name) = $obj->{'name'};
  $name = "default" unless defined($name);
  my($result) = parse_content($obj);
  $lcdml_strings{$name} = $result;
  printDebug(5,"      string name='$name' -> '$result'\n");
}

# ---------- parse_content ----------------------------------------------------
# parse "pos" field and return ($x,$y)
sub parse_pos {
  my($obj) = shift;
  my($pos) = $obj->{'pos'};
  $pos = "0,0" unless defined($pos);
  return split(/,/,$pos);
}

# parse "content" "type" and "filter" and return evaluated string
sub parse_content {
  my($obj) = shift;
  
  # get parameters
  my($content) = $obj->{'content'};
  my($type)    = $obj->{'type'};
  my($filter)  = $obj->{'filter'};
  my($keeplf)   = $obj->{'keeplf'};
  $content = "" unless defined($content);
  $type = "string" unless defined($type);
  $filter = "" unless defined($filter);
  $keeplf = 0 unless defined($keeplf);
  
  # create result
  my($haslf) = 0;
  my($result) = "";
  if($type eq 'shell') {
    open(FH,"$content|") || die "error opening pipe $content";
    $result = join('',<FH>);
    close(FH);
  }
  elsif($type eq 'file') {
    open(FH,"<$content") || die "error opening file $content";
    $result = join('',<FH>);
    close(FH);
  }
  elsif($type eq 'string') {
    $result = $content;
    # variable subsitution $name ${name}
    $result =~ s/\$([A-Za-z]\w*)/if(defined($lcdml_strings{$1})){$lcdml_strings{$1}}/eg;
    $result =~ s/\$\{([A-Za-z]\w*)\}/if(defined($lcdml_strings{$1})){$lcdml_strings{$1}}/eg;
    $result =~ s/\$\$/\$/g;
    # LF replacement, \x00, \\->\ 
    $result =~ s/\\n/\n/g;
    $result =~ s/\\x(\d\d)/chr($1)/eg;
    $result =~ s/\\\\/\\/g;
    $haslf = ($result =~ m/\n/);
  } 
  else {
    print STDERR "Unknown content type: $type\n";
  }

  # filter
  if(!$keeplf && !$haslf) {
    chomp $result;
    $result =~ s/\n/ /g;
  }
  if($filter ne "") {
    eval "\$result =~ s/$filter/\$1/;";
  }
  
  printDebug(6,"\t[content='$content',type='$type',filter='$filter',keeplf=$keeplf->'$result']\n");  
  return $result;
}

# ---------- helper functions -------------------------------------------------
# push_one_or_more(\@array,$ref)
# if $ref is an array add its items otherwise add itself
sub push_one_or_more {
  my($array) = shift;
  my($r) = shift;
  if(ref($r) eq 'ARRAY') {
    foreach(@$r) { push @$array,$_; }
  } else {
    push @$array,$r;
  }
}

sub printDebug {
  my($level) = shift;
  my($string) = shift;
  if($opt_d>=$level) {
    print $string;
  }
}
